/*
 * Copyright (c) 2011-2013 Alexandre Tuleu <alexandre.tuleu.2005@polytechnique.org>
 * Ecole Polytechnique Fédérale de Lausanne.
 *
 * this file is part of sbcpd.
 *
 * sbcpd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * sbcpd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the Lesser GNU General Public License
 * along with sbcpd.  If not, see <http://www.gnu.org/licenses/>.
 */
/**
 * \file bus.c
 *
 *  \date Aug 15, 2011
 *  \author Alexandre Tuleu
 */

#include <stdlib.h>
#include <string.h>

#ifndef NDEBUG
#include <stdio.h>
#endif

#include "bus.ih"
#include <ftdi.h>

#include "transfer.h"

#include "sbcp.h"
#include "error.h"

#include <sbcpd/private/stats.h>


sbcpd_bus_error_code sbcpd_bus_get_last_error_code(sbcpd_bus * bus){
	return (sbcpd_bus_error_code) bus->error.code;
}

const char * sbcpd_bus_get_last_error_message(sbcpd_bus * bus){
	return bus->error.msg;
}

sbcpd_bus * sbcpd_create_bus(const sbcpd_bus_settings * settings){
	sbcpd_bus  * bus = malloc(sizeof(sbcpd_bus));

	sbcpd_init_error(&bus->error,SBCPD_BUS_ERROR);
	bus->flags = SBCPD_BUS_VOID;
	*((sbcpd_frame_size_t  *)&bus->max_frame_packets) = 0;
	*((sbcpd_packet_size_t *)&bus->max_packet_size)   = 0;
	bus->si         = NULL;

	bus->in_buffer  = NULL;
	bus->out_buffer = NULL;
	bus->out_buffer_actual_size = 0;
	bus->stats = NULL;
	*((unsigned int *)&bus->buffers_size) = 0;


	bus->settings= settings;

	//printf("Created a new sbcpd_bus object. \n");

	return bus;
}

#define CHECK_IF_BUS_ALLOCATED if(!bus){\
SBCPD_RETURN_WITH_ERROR(bus,SBCPD_EBUS_NOT_ALLOCATED,\
                        "You passed a NULL pointer !");\
}

#define CHECK_IF_BUS_OPENED(m)if(!(bus->flags & SBCPD_BUS_OPEN)){\
	SBCPD_RETURN_WITH_ERROR(bus,SBCPD_EBUS_IS_NOT_OPEN,\
	                        m);\
}

sbcpd_error_code sbcpd_bus_open(sbcpd_bus * bus,
                   sbcpd_serial_interface * interface){
	CHECK_IF_BUS_ALLOCATED

	if(!interface){
		SBCPD_RETURN_WITH_ERROR(bus,SBCPD_EBUS_SERIAL_INTERFACE_UNVALID,
		                        "serial interface seems unlocatted.");
	}

	if (bus -> si || (bus->flags & SBCPD_BUS_OPEN)){
		SBCPD_RETURN_WITH_ERROR(bus,SBCPD_EBUS_ALREADY_OPEN,
		                        "Bus is already opened, you cannot open"
		                        " it twice !");
	}

	bus->si = interface;

	int res = sbcpd_bus_get_bus_size_and_packet_from_master(bus);

	if(bus->max_frame_packets < bus->settings->minimal_frame_size || bus->max_packet_size < bus->settings->minimal_packet_size){
		static char error_msg_while_opening[800];
		if(res == 0){
			sprintf(error_msg_while_opening,
			        "Succesfully found master, but unmet frame definition : wanted ( %d, %d ), but got ( %d , %d )",
			        bus->settings->minimal_frame_size,
			        bus->settings->minimal_packet_size,
			        bus->max_frame_packets,
			        bus->max_packet_size);
		} else {
			sprintf(error_msg_while_opening,"Unsuccessful opening of master : %s.\n Therefore frame definition is (%d,%d) and we asked for (%d,%d)",
			        bus->error.msg,
			        bus->max_frame_packets,
			        bus->max_packet_size,
			        bus->settings->minimal_frame_size,
			        bus->settings->minimal_packet_size
			        );
		}
		SBCPD_RETURN_WITH_ERROR(bus, SBCPD_EBUS_MINIMAL_FRAME_SIZE_UNMET,
		                        error_msg_while_opening);
	}

	*((unsigned int*)&(bus->buffers_size)) = bus->max_frame_packets
	                                         * bus->max_packet_size;

	bus->out_buffer = malloc(sizeof(sbcpd_byte_t) * bus->buffers_size);
	bus->in_buffer  = malloc(sizeof(sbcpd_byte_t) * bus->buffers_size);
	bus->out_buffer_actual_size = 0;


	bus->transfer_manager = malloc(sizeof(sbcpd_transfer_manager));
	sbcpd_tm_init(bus->transfer_manager,bus->max_frame_packets);

	bus->flags |= SBCPD_BUS_OPEN;

	if(bus->settings->monitor) {
		if (bus->stats){
			sbcpd_free_stats(bus->stats);
		}
		bus->stats = sbcpd_create_stats();
	}

	return SBCPD_NO_ERROR;

}

sbcpd_error_code sbcpd_bus_close(sbcpd_bus * bus){
	CHECK_IF_BUS_ALLOCATED

	if(!(bus->flags & SBCPD_BUS_OPEN)){
		SBCPD_RETURN_WITH_ERROR(bus,
		                        SBCPD_EBUS_IS_NOT_OPEN,
		                        "Bus is not opened, cannot close it");
	}

	//unlink all pending transfer.
	if(bus->flags & SBCPD_BUS_BUSY){
		sbcpd_bus_release_all_transfers(bus);
	}

	if(bus->si){
		(*(bus->si->close))(bus->si);
		(*(bus->si->free))(bus->si);
		bus->si = NULL;
	}

	if(bus->out_buffer){
		free(bus->out_buffer);
	}

	if(bus->in_buffer){
		free(bus->in_buffer);
	}

	*((unsigned int*)&bus->buffers_size) = 0;
	bus->out_buffer_actual_size = 0;

	if(bus->transfer_manager){
		sbcpd_tm_free(bus->transfer_manager);
		free(bus->transfer_manager);
		bus->transfer_manager = 0;
	}

	bus->flags &= (unsigned int)~SBCPD_BUS_OPEN;

	return SBCPD_NO_ERROR;
}

sbcpd_error_code sbcpd_free_bus(sbcpd_bus * bus){
	CHECK_IF_BUS_ALLOCATED

	if(bus->flags & SBCPD_BUS_OPEN){
		sbcpd_bus_close(bus);
	}

	if(bus->stats){
		sbcpd_free_stats(bus->stats);
	}

	free(bus);

	return SBCPD_NO_ERROR;
}

sbcpd_transfer * sbcpd_bus_append_transfer(sbcpd_bus * bus,
                                           unsigned int payload_size){
	if(!bus){
		SBCPD_RETURN_WITH_ERROR_AND_VALUE(bus,
		                                  SBCPD_EBUS_NOT_ALLOCATED,
		                                  "You passed a NULL pointer for a bus",
		                                  NULL);
	}

	if(!(bus->flags & SBCPD_BUS_OPEN)){
		SBCPD_RETURN_WITH_ERROR_AND_VALUE(bus,
		                                  SBCPD_EBUS_IS_NOT_OPEN,
		                                  "You should open the bus to make transfers.",
		                                  NULL);
	}

	if((bus->flags & SBCPD_BUS_BUSY)){
		SBCPD_RETURN_WITH_ERROR_AND_VALUE(bus,
		                                  SBCPD_EBUS_IS_BUSY,
		                                  "You cannot create new transfer while "
		                                  "transfering, please wait for other transfer to "
		                                  "finish",
		                                  NULL);
	}

	if(bus->flags & SBCPD_BUS_FRAME_COMPROMISED){
		SBCPD_RETURN_WITH_ERROR_AND_VALUE(bus,
		                                  SBCPD_EBUS_FRAME_COMPROMISED,
		                                  "The current bus frame is "
		                                  "compromised. Please clear all the "
		                                  "pending transfer before appending a "
		                                  "new one",
		                                  NULL);
	}

	if(payload_size > bus->max_packet_size - SBCP_MINIMAL_SIZE){
		//got some error, I release it
		SBCPD_RETURN_WITH_ERROR_AND_VALUE(bus,
		                                  SBCPD_EBUS_BAD_TRANSFER_DEFINITION,
		                                  "payload size should be between 0 and "
		                                  "bus->max_packet_size - 6",
		                                  NULL);
	}


	sbcpd_transfer * t = sbcpd_tm_allocate_transfer(bus->transfer_manager);


	//no slot available
	if(t == 0){
		SBCPD_RETURN_WITH_ERROR_AND_VALUE(bus,
		                                  SBCPD_EBUS_BUS_RESSOUCE_ALREADY_TOOK,
		                                  "Not enough packet available in the frame",
		                                  NULL);
	}

	t->status = SBCPD_TRANSFER_UNSTARTED;
	t->callback = NULL;
	*((sbcpd_byte_t **)&(t->out_buffer)) = &(bus->out_buffer[bus->out_buffer_actual_size]);
	*((unsigned int *)&t->out_size) = payload_size + SBCP_MINIMAL_SIZE;
	bus->out_buffer_actual_size    += t->out_size;

	*((sbcpd_byte_t**)&t->in_buffer)   = NULL;
	*((unsigned int*)&t->in_size) = 0;

	return t;
}


sbcpd_error_code sbcpd_bus_release_transfer(sbcpd_bus * bus, sbcpd_transfer * t){
	CHECK_IF_BUS_ALLOCATED;
	CHECK_IF_BUS_OPENED("Bus should be opened");

	if(sbcpd_tm_is_allocated(bus->transfer_manager,t) == 0){
		SBCPD_RETURN_WITH_ERROR(bus,SBCPD_EBUS_TRANSFER_NOT_ALLOCATED,
		                        "Transfer is not allocated.");
	}

	if(~(bus->flags & SBCPD_BUS_BUSY)){//transfer is finsihed
		sbcpd_tm_release_transfer(bus->transfer_manager,t);
	} else {
		switch (t->status){
		case SBCPD_TRANSFER_UNSTARTED :
		case SBCPD_TRANSFER_IN_PROGRESS :
			t->status = SBCPD_TRANSFER_ABORTED;
			break;
		case SBCPD_TRANSFER_ABORTED :
			break;
		default :
			sbcpd_tm_release_transfer(bus->transfer_manager,t);
			break;
		}
	}

	//now perform some check
	//check if there is no transfer pending
	if(sbcpd_tm_first_allocated_transfer(bus->transfer_manager) == NULL){
		//reset the frame as there is no transfer pending anymore.
		sbcpd_bus_reset_frame(bus);
	} else {
		//we release one transfer, and the list is not empty.
		//the frame is compromised, we forbid appending and submission.
		bus->flags |= SBCPD_BUS_FRAME_COMPROMISED;
	}

	return SBCPD_NO_ERROR;
}


sbcpd_error_code sbcpd_bus_release_all_transfers(sbcpd_bus * bus){
	CHECK_IF_BUS_ALLOCATED
	CHECK_IF_BUS_OPENED("Bus should be opened.");
	if(sbcpd_tm_first_allocated_transfer(bus->transfer_manager) == NULL){
		SBCPD_RETURN_WITH_ERROR(bus,
		                        SBCPD_EBUS_TRANSFER_LIST_EMPTY,
		                        "Need at least one transfer to remove.");
	}

	sbcpd_transfer * t = sbcpd_tm_first_allocated_transfer(bus->transfer_manager);
	while( t!= NULL ){
		sbcpd_transfer * toDel = t;
		t = sbcpd_tm_next_allocated_transfer(bus->transfer_manager,t);
		sbcpd_bus_release_transfer(bus,toDel);
	}
	return SBCPD_NO_ERROR;
}


sbcpd_error_code sbcpd_bus_submit_all_transfer(sbcpd_bus * bus){
	CHECK_IF_BUS_ALLOCATED
	CHECK_IF_BUS_OPENED("Bus should be opened.");

	if(bus->flags & SBCPD_BUS_FRAME_COMPROMISED){
		SBCPD_RETURN_WITH_ERROR(bus,
		                        SBCPD_EBUS_FRAME_COMPROMISED,
		                        "Transfer frame is being compromised. You "
		                        "should clear all transfers before submission.");
	}

	if(sbcpd_tm_first_allocated_transfer(bus->transfer_manager) == NULL ){
		SBCPD_RETURN_WITH_ERROR(bus,SBCPD_EBUS_TRANSFER_LIST_EMPTY,
		                        "Need at least one transfer definition to send "
		                        "data over the bus");
	}

	if(bus->flags & SBCPD_BUS_BUSY){
		SBCPD_RETURN_WITH_ERROR(bus,SBCPD_EBUS_IS_BUSY,
		                        "Bus is already sending data, wait for current"
		                        "transfer termination.");
	}

	bus->flags |= SBCPD_BUS_BUSY;


	// Preparing packet to send
	sbcpd_transfer * t;
	for(t = sbcpd_tm_first_allocated_transfer(bus->transfer_manager);
	    t != NULL;
	    t = sbcpd_tm_next_allocated_transfer(bus->transfer_manager,t)){
//		if(t.callback == NULL){
//			bus->flags &= ~SBCPD_BUS_BUSY;
//			SBCPD_RETURN_WITH_ERROR(bus,
//			                        SBCPD_EBUS_BAD_TRANSFER_DEFINITION,
//			                        "Transfer does not have a callback set.");
//		}

		//put the correct header
		t->out_buffer[SBCP_HEADER_POS] = SBCP_HEADER_VALUE;

		//computes the right size
		sbcpd_byte_t payload_size = t->out_size - SBCP_MINIMAL_SIZE;
		t->out_buffer[SBCP_PAYLOAD_SIZE_POS] = payload_size;

		//computes the right checksum
		t->out_buffer[payload_size + SBCP_PAYLOAD_START_POS] =
			sbcpd_compute_checksum_of_packet(t->out_buffer);
	}
	//all packet prepared successfully.

	//setting all the transfer status
	for(t = sbcpd_tm_first_allocated_transfer(bus->transfer_manager);
	    t != NULL;
	    t = sbcpd_tm_next_allocated_transfer(bus->transfer_manager,t)){
		if(t->status != SBCPD_TRANSFER_ABORTED){
			t->status = SBCPD_TRANSFER_IN_PROGRESS;
		}
	}


	//sending all the data over the host bus.
	int sended_bytes = (*(bus->si->write_data))(bus->si,
	                                            bus->out_buffer,
	                                            (int)bus->out_buffer_actual_size);

	if( sended_bytes < 0 ){
		for(t = sbcpd_tm_first_allocated_transfer(bus->transfer_manager);
			    t != NULL;
			    t = sbcpd_tm_next_allocated_transfer(bus->transfer_manager,t)){
			t->status = SBCPD_TRANSFER_ABORTED;
		}
		bus->flags &= (unsigned int)~SBCPD_BUS_BUSY;
		SBCPD_RETURN_WITH_ERROR(bus,SBCPD_EBUS_CANNOT_SEND_DATA,
		                        "serial interface was not able to send data.");
	}

	//now get all the response packet.
	sbcpd_bus_get_response_packets(bus);
	//all packet are received, the bus is not busy anymore.
	bus->flags &= (unsigned int)~SBCPD_BUS_BUSY;
	return SBCPD_NO_ERROR;
}


sbcpd_error_code sbcpd_bus_wait_for_completion(sbcpd_bus * bus, int timeout __attribute__((unused))){
	CHECK_IF_BUS_ALLOCATED;
	CHECK_IF_BUS_OPENED("Bus should be opened.")

	if(bus->flags & SBCPD_BUS_FRAME_COMPROMISED){
		SBCPD_RETURN_WITH_ERROR(bus,
		                        SBCPD_EBUS_FRAME_COMPROMISED,
		                        "Frame is compromised, please clear all "
		                        "transfers.");
	}
	return SBCPD_NO_ERROR;
}


int sbcpd_bus_good(sbcpd_bus * bus){
	if( bus == NULL ){
		return 0;
	}

	if( (bus->flags & SBCPD_BUS_OPEN) == 0 ){
		SBCPD_RETURN_WITH_ERROR_AND_VALUE(bus,SBCPD_EBUS_IS_NOT_OPEN,
		                                  "Bus is not openened",0);
	}

	if( bus->flags & SBCPD_BUS_FRAME_COMPROMISED ){
		SBCPD_RETURN_WITH_ERROR_AND_VALUE(bus,SBCPD_EBUS_FRAME_COMPROMISED,
		                                  "Frame is compromised.",0);
	}

	return 1;
}

int sbcpd_bus_is_busy(sbcpd_bus * bus){
	if(bus == NULL){
		return 0;
	}

	return ((bus->flags & SBCPD_BUS_BUSY) != 0 ) ? 1 : 0;
}

sbcpd_frame_size_t sbcpd_bus_max_packet_per_frame(sbcpd_bus * bus){
	return bus->max_frame_packets;
}

sbcpd_packet_size_t sbcpd_bus_max_packet_size(sbcpd_bus * bus){
	return bus->max_packet_size;
}


sbcpd_frame_size_t sbcpd_bus_current_frame_size(sbcpd_bus * bus){
	if( bus == NULL){
		return 0;
	}
	return sbcpd_tm_transfer_count(bus->transfer_manager);
}

int sbcpd_bus_frame_is_empty(sbcpd_bus * bus){
	if(bus == NULL){
		return 1;
	}
	return sbcpd_tm_first_allocated_transfer(bus->transfer_manager) == NULL;
}

int sbcpd_bus_read_needed_byte(sbcpd_bus * bus,
                               unsigned int needed_bytes){
	unsigned int maxTrials = bus->settings->max_trials;
	while(   bus->received_bytes < bus->parsed_bytes + needed_bytes
	      && maxTrials > 0){
		int readout =
			(*bus->si->read_data)(bus->si,
			                      &(bus->in_buffer[bus->received_bytes]),
		                          (int)(bus->buffers_size - bus->received_bytes));

		if(readout < 0){ //report a fatl readout error.
			maxTrials = 0;
			break;
		}

		bus->received_bytes += (unsigned int)readout;
		--maxTrials;
	}

	if(maxTrials == 0 && bus->received_bytes < bus->parsed_bytes + needed_bytes){
//		for(; current_packet < bus->transfers_size; ++current_packet){
//			bus->transfers[current_packet].com_error =
//				SBCPD_CERR_IMCOMPLETE_READOUT;
//			bus->transfers[current_packet].status =
//				SBCPD_TRANSFER_HAS_COMMUNICATION_ERROR;
//		}
		return 1;
	}
	return 0;
}



sbcpd_byte_t sbcpd_compute_checksum_of_packet(const sbcpd_byte_t * data){
	sbcpd_byte_t cs = 0;
	unsigned int i = SBCP_CLASS_POS;
	sbcpd_byte_t cs_pos = data[SBCP_PAYLOAD_SIZE_POS] + SBCP_PAYLOAD_START_POS;
	for(;i < cs_pos; ++i){
		cs += data[i];
	}
	return cs;
}

void sbcpd_transfer_parse_communication_errors(sbcpd_transfer * t){

	if(t->in_size != 6){
		t->com_error = SBCPD_CERR_BAD_CERR_PACKET;
		return;
	}

	switch(t->in_buffer[SBCP_INSTRUCTION_POS]){
	case SBCP_CERR_CS_INCORRECT :
		if(t->in_buffer[SBCP_CLASS_POS] == SBCP_MASTER_CLASS){
			t->com_error = SBCPD_CERR_HOST_BAD_CHECKSUM;
		} else {
			t->com_error = SBCPD_CERR_MASTER_BAD_CHECKSUM;
		}
		break;
	case SBCP_CERR_BUS_TIMEOUT :
		t->com_error = SBCPD_CERR_HOST_TIMEDOUT;
		break;
	case SBCP_CERR_FRAME_ERROR :
		if(t->in_buffer[SBCP_CLASS_POS] == SBCP_MASTER_CLASS){
			t->com_error = SBCPD_CERR_HOST_FRAME_ERROR;
		} else {
			t->com_error = SBCPD_CERR_MASTER_FRAME_ERROR;
		}
	case SBCP_CERR_SLAVE_TIMEOUT :
		if(t->in_buffer[SBCP_CLASS_POS] == SBCP_MASTER_CLASS){
			t->com_error = SBCPD_CERR_SLAVE_TIMEOUT;
		}else{
			t->com_error = SBCPD_CERR_MASTER_TIMEDOUT;
		}
		break;
	case SBCP_CERR_HOST_FRAME_ERROR :
		t->com_error = SBCPD_CERR_HOST_FRAME_ERROR;
		break;
	case SBCP_CERR_SLAVE_FRAME_ERROR :
		if(t->in_buffer[SBCP_CLASS_POS] == SBCP_MASTER_CLASS){
			t->com_error = SBCPD_CERR_SLAVE_FRAME_ERROR;
		} else {
			t->com_error = SBCPD_CERR_MASTER_FRAME_ERROR;
		}
		break;
	case SBCP_CERR_UNKNOWN_INSTRUCTION :
		t->com_error = SBCPD_CERR_UNKNOWN_INSTRUCTION_ERROR;
		break;
	case SBCP_CERR_UNEXPECTED_SIZE :
		t->com_error = SBCP_CERR_UNEXPECTED_SIZE;
		break;
	case SBCP_CERR_REG_NOT_READABLE :
		t->com_error = SBCPD_CERR_REG_NOT_READABLE;
		break;
	case SBCP_CERR_REG_NOT_WRITABLE :
		t->com_error = SBCPD_CERR_REG_NOT_WRITABLE;
		break;
	case SBCP_CERR_REG_PROTECTED :
		t->com_error = SBCPD_CERR_REG_PROTECTED;
		break;
	default :
		t->com_error = SBCPD_CERR_BAD_CERR_PACKET;
		break;
	}
}


void sbcpd_transfer_check_response(sbcpd_transfer * t){
	sbcpd_byte_t cs = sbcpd_compute_checksum_of_packet(t->in_buffer);

	if(cs != t->in_buffer[t->in_size - 1] ){
		t->com_error = SBCPD_CERR_SLAVE_BAD_CHECKSUM;
		t->status    = SBCPD_TRANSFER_HAS_COMMUNICATION_ERROR;
		return;
	}

	if(t->in_buffer[SBCP_INSTRUCTION_POS] & SBCP_COMMUNICATION_ERROR_BIT){
		t->status = SBCPD_TRANSFER_HAS_COMMUNICATION_ERROR;
		sbcpd_transfer_parse_communication_errors(t);
		return;
	}

	if(t->in_buffer[SBCP_INSTRUCTION_POS] & SBCP_APPLICATION_ERROR_MASK){
		t->status = SBCPD_TRANSFER_HAS_APPLICATION_ERROR;
		return;
	}

	t->status = SBCPD_TRANSFER_OK;
	//we got an error to parse.

}

int sbcpd_bus_get_one_packet_data(sbcpd_bus * bus, sbcpd_transfer * transfer){

	do {
		if(sbcpd_bus_read_needed_byte(bus,
		                              SBCP_HEADER_POS + 1)){
			//fatal error for the batch
			return 1;
		}
	}while(bus->in_buffer[bus->parsed_bytes] != SBCP_HEADER_VALUE &&
			++bus->parsed_bytes);

	if(sbcpd_bus_read_needed_byte(bus,
	                              SBCP_PAYLOAD_SIZE_POS + 1)){
		//fatal error for the batch
		return 1;
	}

	unsigned int packet_total_size =
					bus->in_buffer[bus->parsed_bytes + SBCP_PAYLOAD_SIZE_POS]
					               + SBCP_MINIMAL_SIZE;

	if(packet_total_size > bus->max_packet_size){
		transfer->status    = SBCPD_TRANSFER_HAS_COMMUNICATION_ERROR;
		transfer->com_error = SBCPD_CERR_INCONSISTENT_RESPONSE_SIZE;
		//ignore the bad started header, but allow the rest of the frame
		//to be parsed
		bus->parsed_bytes += SBCP_PAYLOAD_SIZE_POS + 1;
		return 0;//not a fatal error.
	}

	if(sbcpd_bus_read_needed_byte(bus,packet_total_size)){
		//fatal error.
		return 1;
	}

	*((sbcpd_byte_t **)&transfer->in_buffer) = &(bus->in_buffer[bus->parsed_bytes]);
	*((unsigned int*)&transfer->in_size)  = packet_total_size;
	bus->parsed_bytes += packet_total_size;
	return 0;
}

void sbcpd_bus_get_response_packets(sbcpd_bus * bus){
	bus->parsed_bytes   = 0;
	bus->received_bytes = 0;

	//crashed store if the frame data is ok. if a fatal parsing error occurs,
	//we should stop reading
	unsigned int crashed = 0;
	sbcpd_transfer * t;
	for(t = sbcpd_tm_first_allocated_transfer(bus->transfer_manager);
		t != NULL;
		t = sbcpd_tm_next_allocated_transfer(bus->transfer_manager,t)){
		//getting the data from interface
		//if frame is fatally compromised, we don't read any further.
		if(crashed == 0){
			crashed = sbcpd_bus_get_one_packet_data(bus,t) != 0;
		}
		//now we check if the packet has been aborted
		//we should do this check after receiving the packet, in order not to mess
		//the frame reading (abortion of a packet don't prevent it to be send)
		//only the callback to be discarded.
		if(t->status == SBCPD_TRANSFER_ABORTED){
			//we should release it now.
			sbcpd_tm_release_transfer(bus->transfer_manager,t);
			continue;
		}

		if(crashed == 0){
			//data is fine, we check for communication errors
			sbcpd_transfer_check_response(t);
		} else {//frame has been compromised, we mark packets as incomplete.
			t->status = SBCPD_TRANSFER_HAS_COMMUNICATION_ERROR;
			t->com_error = SBCPD_CERR_IMCOMPLETE_READOUT;
		}

		if(bus->stats){
			sbcpd_stats_log_transfer(bus->stats,t);
		}

		if(t->callback){
			//we need to call the callback, no matter what !
			(*(t->callback))(t);
		}
	}
}


void sbcpd_bus_reset_frame(sbcpd_bus * bus){
	//we clear the compromised flags
	bus->flags &= (unsigned int)~(SBCPD_BUS_FRAME_COMPROMISED);
	bus->out_buffer_actual_size = 0;
}

#define SET_BUS_PACKET_FRAME_VALUE(bus,packetSize,frameSize)do{\
	*((sbcpd_frame_size_t  *)&bus->max_frame_packets) = frameSize;\
	*((sbcpd_packet_size_t *)&bus->max_packet_size)   = packetSize;\
}while(0)


#define RETURN_WITH_NO_MASTER(bus, ...)do{\
	SET_BUS_PACKET_FRAME_VALUE(bus,255,1);\
	sprintf(err_on_master_opening , __VA_ARGS__);\
	bus->error.msg = err_on_master_opening; \
	return 1;\
}while(0)

int send_packet_and_get_response(sbcpd_bus *bus,
                                 sbcpd_packet_size_t outSize,
                                 const sbcpd_byte_t * out,
                                 sbcpd_packet_size_t expectedResponseSize,
                                 sbcpd_byte_t * in){
	int sended_bytes = (*(bus->si->write_data))(bus->si,
	                                            out,
	                                            outSize);

	if(sended_bytes != outSize){
		return 1;
	}

	sbcpd_byte_t          maxTrial = 20;
	sbcpd_packet_size_t readenByte = 0;
	while(maxTrial > 0 && readenByte < expectedResponseSize){
		int read = (*(bus->si->read_data))(bus->si,in,261);
		if(read < 0){
			break;
		}
		--maxTrial;
		readenByte += read;
	}

	if (readenByte < expectedResponseSize){
		return 2;
	}

	return 0;
}

int sbcpd_bus_get_bus_size_and_packet_from_master(sbcpd_bus * bus){
	sbcpd_byte_t out[8];
	sbcpd_byte_t in[261];
	static char err_on_master_opening[400];
	//generate some get firmware packet
	out[SBCP_HEADER_POS] = SBCP_HEADER_VALUE;
	out[SBCP_CLASS_POS]  = SBCP_MASTER_CLASS;
	out[SBCP_ID_POS ]    = SBCP_MASTER_BROADCAST_ID;
	out[SBCP_INSTRUCTION_POS] = SBCP_GET_FIRMWARE_INSTRUCTION; //get firmware
	out[SBCP_PAYLOAD_SIZE_POS] = 0x00;
	out[SBCP_PAYLOAD_START_POS] = sbcpd_compute_checksum_of_packet(out);

	if(send_packet_and_get_response(bus,6,out,8,in)){
		RETURN_WITH_NO_MASTER(bus , "could not get firmware code from master\n");
	}

	sbcpd_packet_size_t size = in[SBCP_PAYLOAD_SIZE_POS];
	if(size % 2 != 0){
		RETURN_WITH_NO_MASTER(bus, "Unexpected size %d (expected %d) while getting firmware\n",size,2);
	}

	if(in[SBCP_INSTRUCTION_POS] != SBCP_NO_ERROR){
		RETURN_WITH_NO_MASTER(bus , "Got error %d while getting firmware\n",in[SBCP_INSTRUCTION_POS]);
	}

	if(in[SBCP_PAYLOAD_START_POS + size] != sbcpd_compute_checksum_of_packet(in)){
		RETURN_WITH_NO_MASTER(bus, "Unmatched cs on getting firmware\n");
	}

	out[SBCP_INSTRUCTION_POS] = SBCP_GET_REGISTER_INSTRUCTION;
	out[SBCP_PAYLOAD_SIZE_POS] = 0x02;
	out[SBCP_PAYLOAD_START_POS] = SBCP_MASTER_FRAME_SIZE_REG_ADDRESS;
	out[SBCP_PAYLOAD_START_POS + 1] = SBCP_MASTER_PACKET_SIZE_REG_ADDRESS;
	out[SBCP_PAYLOAD_START_POS + 2] = sbcpd_compute_checksum_of_packet(out);

	if(send_packet_and_get_response(bus,8,out,10,in)){
		RETURN_WITH_NO_MASTER(bus, "could not get packet definition from master\n");
	}

	sbcpd_frame_size_t frame_size;
	sbcpd_packet_size_t packet_size;
	size = in[SBCP_PAYLOAD_SIZE_POS];
	if(size != 4){
		RETURN_WITH_NO_MASTER(bus, "Unexpected size %d (expected %d) while getting packet def.\n",size,4);
	}

	if(in[SBCP_INSTRUCTION_POS] != SBCP_NO_ERROR){
		RETURN_WITH_NO_MASTER(bus, "Got error %d while getting packet def\n",in[SBCP_INSTRUCTION_POS]);
	}


	if(in[SBCP_PAYLOAD_START_POS + size] != sbcpd_compute_checksum_of_packet(in)){
		RETURN_WITH_NO_MASTER(bus,"Unmatched cs (got 0x%x , expected 0x%x )on getting packet definition.\n",
		                      in[SBCP_PAYLOAD_START_POS + size],
		                      sbcpd_compute_checksum_of_packet(in));
	}

	frame_size   = in[SBCP_PAYLOAD_START_POS + 1];
	frame_size  += ((uint16_t)in[SBCP_PAYLOAD_START_POS]) << 8;
	packet_size  = in[SBCP_PAYLOAD_START_POS + 3];
	packet_size += ((uint16_t)in[SBCP_PAYLOAD_START_POS + 2]) << 8;

	SET_BUS_PACKET_FRAME_VALUE(bus,packet_size,frame_size);
	return 0;
}

sbcpd_stats * sbcpd_bus_get_stats(sbcpd_bus * bus){
	if(!bus){
		SBCPD_RETURN_WITH_ERROR_AND_VALUE(bus,
		                                  SBCPD_EBUS_NOT_ALLOCATED,
		                                  "You passed a NULL pointer, please sbcpd_create_bus()",
		                                  NULL);
	}

	if((bus->flags & SBCPD_BUS_OPEN) == 0){
		SBCPD_RETURN_WITH_ERROR_AND_VALUE(bus,
		                                  SBCPD_EBUS_IS_NOT_OPEN,
		                                  "Bus is not yet opened.",
		                                  NULL);
	}

	if(bus->stats == NULL) {
		SBCPD_RETURN_WITH_ERROR_AND_VALUE(bus,
		                                  SBCPD_EBUS_MONITORING_DISABLED,
		                                  "Monitoring is disabled on the bus, please enable it in settongs before opening.",NULL);
	}

	return bus->stats;
}



