/*
 * Copyright (c) 2011-2013 Alexandre Tuleu <alexandre.tuleu.2005@polytechnique.org>
 * Ecole Polytechnique Fédérale de Lausanne.
 *
 * this file is part of sbcpd.
 *
 * sbcpd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * sbcpd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the Lesser GNU General Public License
 * along with sbcpd.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * \file bus.h
 *
 *  \date Aug 15, 2011
 *  \author Alexandre Tuleu
 */

#ifndef SBCPD_BUS_H_
#define SBCPD_BUS_H_

#include <stdint.h>
#include "common.h"
#include "error.h"
#include "stats.h"


#ifdef __cplusplus
 extern "C" {
 #endif



/**
 * Error code that could be returned by the \ref sbcpd_bus_m  'sfunctions.
 * \ingroup sbcpd_bus_m
 */
enum sbcpd_bus_error_code_enm {
	SBCPD_EBUS_NO_ERROR                      = 0x0000,
	//!< SBCPD_EBUS_NO_ERROR No errors happened.
	SBCPD_EBUS_ALREADY_OPEN,
	//!< SBCPD_EBUS_ALREADY_OPENED Bus is already opened.
	SBCPD_EBUS_NOT_ALLOCATED,
	//!< SBCPD_EBUS_NOT_ALLOCATED the bus is not allocated
	SBCPD_EBUS_IS_NOT_OPEN,
	//!< SBCPD_EBUS_IS_NOT_OPEN the bus was not opened when running the command.
	SBCPD_EBUS_BUS_RESSOUCE_ALREADY_TOOK,
	//!< SBCPD_EBUS_IS_ALREADY_OPENED the bus ressources is already allocated
	//!< to another application.
	SBCPD_EBUS_IS_BUSY,
	//!< SBCPD_EBUS_IS_BUSY The bus is currently busy, the operation could not
	//!< be performed.
	SBCPD_EBUS_TRANSFER_LIST_EMPTY,
	//!< SBCPD_EBUS_TRAMNSFER_LIST_EMPTY There is no transfer definition pending
	//!< for this bus.
	SBCPD_EBUS_BAD_TRANSFER_DEFINITION,
	//!< SBCPD_EBUS_BAD_TRANSFER_DEFINITION The transfer definition isn't
	//!< correct.
	SBCPD_EBUS_CANNOT_SEND_DATA,
	SBCPD_EBUS_TRANSFER_NOT_ALLOCATED,
	SBCPD_EBUS_FRAME_COMPROMISED,
	SBCPD_EBUS_MINIMAL_FRAME_SIZE_UNMET,
	SBCPD_EBUS_SERIAL_INTERFACE_UNVALID,
	SBCPD_EBUS_MONITORING_DISABLED,
};

typedef enum sbcpd_bus_error_code_enm  sbcpd_bus_error_code;




/**
 * Represents a SBCP bus that could be issued to manipulate sbcpd_transfer.
 * End user should not write in any of the structure fields, at the exception of
 * the settings, before opening.
 * \note: The sbcp_bus get the ownership of the settings.
 * \ingroup sbcpd_bus_m
 */

typedef struct sbcpd_bus sbcpd_bus;



/**
 * Creates a new bus with given settings. Will own the settings structure.
 * \return
 */
sbcpd_bus * sbcpd_create_bus(const sbcpd_bus_settings  * settings);

/**
 * Finds and opens a bus through a serial interface. Please notice that
 * the settings of the bus should already be set in the sbcpd_bus structure.
 * \ingroup sbcpd_bus_m
 * \note : the sbcpd_bus get the ownership of the interface.
 * \param bus the bus structure with settings already set.
 * \param interface the sbcpd_serial_interface to use.
 * \return SBCPD_NO_ERROR on success, otherwise any of the following code :
 *  - SBCPD_EBUS_NOT_FOUND if the specified device is not found.
 *  - SBCPD_EBUS_ALREADY_OPEN if the specified bus is already opened.
 */

sbcpd_error_code sbcpd_bus_open(sbcpd_bus* bus,
                                sbcpd_serial_interface * interface);


/**
 * Closes a bus. This should be done to release ownership of the application on
 * the bus.
 * \ingroup sbcpd_bus_m
 * \param bus
 * \return SBCPD_NO_ERROR or any of the following error code :
 * - SBCPD_EBUS_NOT_ALLOCATED
 * - SBCPD_EBUS_IS_NOT_OPEN
 */
sbcpd_error_code sbcpd_bus_close(sbcpd_bus * bus);

/**
 * Free the memory used by a bus.
 * \ingroup sbcpd_bus_m
 * \param bus
 * \return SBCPD_NO_ERROR or SBCPD_EBUS_NOT_ALLOCATED if the bus is not
 * allocated.
 */
sbcpd_error_code sbcpd_free_bus(sbcpd_bus * bus);



/**
 * Append a new transfer definition to the bus. memory of the returned structure
 * is managed intrenally and is valid until sbcpd_bus_release_transfer is
 * called.
 * \param bus the sbcpd_bus to use
 * \param out_size the output size of the transfer.
 * \return a pointer to the new sbcpd_transfer structure when there is no error
 *         or NULL otherwise. It may set the following bus errors :
 * -
 * \ingroup sbcpd_bus_m
 */
sbcpd_transfer * sbcpd_bus_append_transfer(sbcpd_bus * bus,
                                           unsigned int payload_size);



/**
 * Release the current transfer. If transfer is in progress, callback wuill be
 * ignored. Data pointed by the structure should not be considered valid
 * anymore.
 * \param bus
 * \return SBCPD_NO_ERROR on success, or any of the following error code :
 * - SBCPD_EBUS_NOT_ALLOCATED
 * - SBCPD_EBUS_NOT_OPENED
 * - SBCPD_EBUS_TRANSFER_NOT_ALLOCATED
 * \ingroup sbcpd_bus_m
 */
sbcpd_error_code sbcpd_bus_release_transfer(sbcpd_bus * bus,
                                            sbcpd_transfer * t);

/**
 * Release all the transfer of the bus.
 * \param bus
 * \return SBCPD_NO_ERROR on success, or any of the following error code :
 * - SBCPD_EBUS_NOT_ALLOCATED
 * - SBCPD_EBUS_NOT_OPENED
 * - SBCPD_EBUS_TRANSFER_LIST_EMPTY
 * \ingroup sbcpd_bus_m
 */
sbcpd_error_code sbcpd_bus_release_all_transfers(sbcpd_bus * bus);

/**
 * Submits all the current affected tranfer to the bus.
 * \param bus
 * \return SBCPD_NO_ERROR on success, or any of the following error code :
 * - SBCPD_EBUS_NOT_ALLOCATED
 * - SBCPD_EBUS_NOT_OPENED
 * - SBCPD_EBUS_TRANSFER_LIST_EMPTY : if no transfer were asked.
 * - SBCPD_EBUS_FRAME_COMPROMISED : if the io frame is compromised. You should
 * then release all the pending transfers to create a new empty frame.
 * \ingroup sbcpd_bus_m
 */
sbcpd_error_code sbcpd_bus_submit_all_transfer(sbcpd_bus * bus);

/**
 * Blocks current thread until completion of all pending bus I/O,
 * or timeout expired.
 * @param bus the bus to wait
 * @param timeout \leq 0, wait for an infinite time, otherwise issue a timeout
 *        error.
 * @return SBCPD_NO_ERROR or any of the following errors :
 * - SBCPD_EBUS_NOT_ALLOCATED
 * - SBCPD_EBUS_NOT_OPENED
 * - SBCPD_EBUS_FRAME_COMPROMISED : if the io frame is compromised. You should
 * then release all the pending transfers to create a new empty frame.
 * \ingroup sbcpd_bus_m
 */
sbcpd_error_code sbcpd_bus_wait_for_completion(sbcpd_bus * bus, int timeout);

/**
 * Checks if the bus is in a good state.
 * @param bus
 * @return 1 if the bus is ok, or 0 otherwise. It sets the bus error with any
 * of the following errors :
 * - SBCPD_EBUS_NOT_OPENED
 * - SBCPD_EBUS_FRAME_COMPROMISED
 */

int sbcpd_bus_good(sbcpd_bus * bus);


/**
 * Checks if the bus is busy
 * @param bus
 * @return 0 if not busy, 1 otherwise.
 */
int sbcpd_bus_is_busy(sbcpd_bus * bus);


/**
 * \ingroup sbcpd_bus_m
 * @param bus
 * @return the size of the bus current frame.
 */
sbcpd_frame_size_t sbcpd_bus_current_frame_size(sbcpd_bus * bus);

/**
 * \ingroup sbcpd_bus_m
 * @param bus
 * @return 1 if the bus frame is empty
 */

int sbcpd_bus_frame_is_empty(sbcpd_bus * bus);



/**
 * \return the last bus error code
 * \ingroup sbcpd_bus_m
 * \param bus
 */

//static inline sbcpd_bus_error_code sbcpd_bus_get_last_error_code(sbcpd_bus * bus){
//	return (sbcpd_bus_error_code) sbcpd_get_error_code(&(bus->error));
//}
sbcpd_bus_error_code sbcpd_bus_get_last_error_code(sbcpd_bus * bus);

//static inline const char * sbcpd_bus_get_last_error_message(sbcpd_bus_str * bus){
//	return bus->error.msg;
//}
const char * sbcpd_bus_get_last_error_message(sbcpd_bus * bus);

/**
 * @param bus the bus tor read from.
 * \ingroup sbcpd_bus_m
 * @return the size of the bus frame
 */
sbcpd_frame_size_t sbcpd_bus_max_packet_per_frame(sbcpd_bus * bus);

/**
 * @param bus the bus to read from.
 * \ingroup sbcpd_bus_m
 * @return the maximal size of a packet (payload and headers ).
 */
sbcpd_packet_size_t sbcpd_bus_max_packet_size(sbcpd_bus * bus);

/**
 * Returns the stats of the bus.
 * \warning : monitoring should have been enabled in settings.
 */
sbcpd_stats * sbcpd_bus_get_stats(sbcpd_bus * bus);


/**
 * \defgroup sbcpd_bus_m SBCP Bus
 *
 *
 * \ingroup sbcpd_m
 *
 */

#ifdef __cplusplus
 }
#endif


#endif //SBCPD_BUS_H_
