/*
 * Copyright (c) 2011-2013 Alexandre Tuleu <alexandre.tuleu.2005@polytechnique.org>
 * Ecole Polytechnique Fédérale de Lausanne.
 *
 * this file is part of sbcpd.
 *
 * sbcpd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * sbcpd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the Lesser GNU General Public License
 * along with sbcpd.  If not, see <http://www.gnu.org/licenses/>.
 */
/**
 * \file error.h
 *
 *  \date Aug 24, 2011
 *  \author Alexandre Tuleu
 */

#ifndef SBCPD_ERROR_H_
#define SBCPD_ERROR_H_

#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef uint32_t sbcpd_error_code;

static const sbcpd_error_code SBCPD_NO_ERROR = 0;

/**
 * Domain of errors that could occurs in the SBCP low level drivers.
 * \ingroup sbcpd_error_m
 */
enum sbcpd_error_domain{
	SBCPD_GLOBAL_ERROR              = 0x0001,
	SBCPD_BUS_ERROR                 = 0x0002,
	SBCPD_SERIAL_INTERFACE_ERROR    = 0x0003,
	SBCPD_TRANSFER_ERROR            = 0x0004
};
typedef enum sbcpd_error_domain sbcpd_error_domain;

/**
 * Represents an error of the sbcp low level driver.
 * \ingroup sbcpd_error_m
 */
struct sbcpd_error_str {
	/**
	 * A code that represents the error. If it is zero, there is no error.
	 */
	sbcpd_error_code  code;

	/**
	 * The domain of the error.
	 */
	sbcpd_error_domain domain;
	/**
	 * Message associated with the error.
	 */
	const char *       msg;
};

typedef struct sbcpd_error_str sbcpd_error;



/**
 * Initializes the error.
 * \ingroup sbcpd_error_m
 * \param error the sbcpd_error to initialize.
 * \param domain the sbcpd_error_domain used for this error.
 */
void sbcpd_init_error(sbcpd_error * error, sbcpd_error_domain domain);

/**
 * Creates a new error.
 * @return a pointer to the new sbcpd_error
 * \ingroup sbcpd_error_m
 */
sbcpd_error * sbcpd_create_global_error();

/**
 * Releases an allocated error.
 * @param error
 * \ingroup sbcpd_error_m
 */
void sbcpd_free_global_error(sbcpd_error * error);



/**
 * \return the error code of an sbcpd_error
 * \param error the error to get the code from.
 * \ingroup sbcpd_error_m
 */
static inline uint32_t sbcpd_get_error_code(sbcpd_error * error){
	return error->code;
}


/**
 * \return the domain of an sbcpd_error
 *
 * \ingroup sbcpd_error_m
 */
static inline sbcpd_error_domain sbcpd_get_error_domain(sbcpd_error * error){
	return error->domain;
}



#define SBCPD_RETURN_WITH_ERROR_AND_VALUE(var,error_code,error_msg,value) do{\
	var->error.code = (uint32_t)error_code;\
	var->error.msg = error_msg;\
	return value;\
}while(0)
/**
 * A convenience macro to return of a function that causes an error.
 * \ingroup sbcpd_error_m
 */
#define SBCPD_RETURN_WITH_ERROR(var, error_code , error_msg) \
	SBCPD_RETURN_WITH_ERROR_AND_VALUE(var,error_code,error_msg,error_code)



/**
 * \defgroup sbcpd_error_m Error Handling
 *
 * In the SBCP low level driver, errors are reported using an sbcpd_error
 * structure. It has an sbcpd_error_domain, which id unique for each module,
 * and an error code, which if it is zero, means that no errors occured.
 * \ingroup sbcpd_m
 */

#ifdef __cplusplus
}
#endif

#endif // SBCPD_ERROR_H_
