/*
 * Copyright (c) 2011-2013 Alexandre Tuleu <alexandre.tuleu.2005@polytechnique.org>
 * Ecole Polytechnique Fédérale de Lausanne.
 *
 * this file is part of sbcpd.
 *
 * sbcpd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * sbcpd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the Lesser GNU General Public License
 * along with sbcpd.  If not, see <http://www.gnu.org/licenses/>.
 */
/**
 * \file ft2232h.c
 *
 *  \date Jul 25, 2011
 *  \author Alexandre Tuleu
 */

#include "ft2232h.h"

#define ftdi_error_return(code, str) do {  \
		ftdi->error_str = str;\
		return code;\
} while(0);

int ft2232h_convert_baudrate(int baudrate,
                             unsigned short *value,
                             unsigned short *index){



	double clock[2] = {3000000,12000000};

	double sub_integers[8] = {0,0.5,0.25,0.125,0.375,0.675,0.75,0.875};

	unsigned int best_clock_index = 2;
	unsigned int best_integer     = 0;
	unsigned int best_sub_integer = 8;
	double best_error = baudrate;
	unsigned int ci;
	for(ci = 0; ci < 2; ++ci){
		double divisor = clock[ci] /  baudrate;
		if(divisor < 1.0){
			continue;
		}
		unsigned int integer = floor(divisor);
		if(integer >= (0x01 << 14)) { // integer  >= 2 ^ 14
			continue;
		}
		unsigned int si;
		for(si = 0; si < 8 ; ++si){
			double reached_baudrate = clock[ci] / ((double)(integer) + sub_integers[si]);
			double error = abs(reached_baudrate - baudrate);
			if(error <= best_error){
				best_error = error;
				best_clock_index = ci;
				best_integer = integer;
				best_sub_integer = si;
			}
		}
	}

	if(best_clock_index == 2){
		return 0;
	}

	*value = best_integer & 0x3fff; //get the 14 first bits only
	*value = ((best_sub_integer & 0x03)<< 14) + *value; //left shift of 14
	*index = ((best_sub_integer & 0x04) >> 2);//only the the last bit
	*index += (best_clock_index << 1); //add clock if needed
	*index = *index << 8; //now format it
	return clock[best_clock_index] / ((double)(best_integer) + sub_integers[best_sub_integer]);
}


int ft2232h_set_baudrate(struct ftdi_context *ftdi,int baudrate){
	unsigned short value,index;
	int reachable_baudrate;

	if(ftdi == NULL || ftdi->usb_dev == NULL){
		ftdi_error_return(-3, (char *)"USB device unavailable");
	}

	if(ftdi->bitbang_enabled){
		baudrate *= 4;
	}

	reachable_baudrate = ft2232h_convert_baudrate(baudrate,&value,&index);

	if(reachable_baudrate == 0){
		ftdi_error_return(-1, (char *)"Unreachable baudrate !");
	}

	//check a 3% boundary
	double margin = ((double)(reachable_baudrate - baudrate))/((double) baudrate);

	if(abs(margin) >= 0.03){
		ftdi_error_return(-1, (char *)"Unsupported baudrate. Note : BitBang baudrate are automatically multiplied by 4.");
	}

	index |= ftdi->index;//don't forget to specify the right interface !!!

	if(usb_control_msg(ftdi->usb_dev,
	                   FTDI_DEVICE_OUT_REQTYPE,
	                   SIO_SET_BAUDRATE_REQUEST,
	                   value,
	                   index,
	                   NULL,
	                   0,
	                   ftdi->usb_write_timeout) != 0){
		ftdi_error_return(-2, (char *)"Setting the new baudrate failed.");
	}
	ftdi->baudrate = reachable_baudrate;
	return 0;
}

int ft2232h_set_latency_timer(struct ftdi_context *ftdi,
                              unsigned char latency){
	unsigned short usb_val;


	if (ftdi == NULL || ftdi->usb_dev == NULL)
		ftdi_error_return(-3, (char *)"USB device unavailable");

	usb_val = latency;
	if (usb_control_msg(ftdi->usb_dev,
	                    FTDI_DEVICE_OUT_REQTYPE,
	                    SIO_SET_LATENCY_TIMER_REQUEST,
	                    usb_val,
	                    ftdi->index,
	                    NULL,
	                    0,
	                    ftdi->usb_write_timeout) != 0)
		ftdi_error_return(-2, (char *)"unable to set latency timer");

	return 0;
}

