/*
 * Copyright (c) 2011-2013 Alexandre Tuleu <alexandre.tuleu.2005@polytechnique.org>
 * Ecole Polytechnique Fédérale de Lausanne.
 *
 * this file is part of sbcpd.
 *
 * sbcpd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * sbcpd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the Lesser GNU General Public License
 * along with sbcpd.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * \file tiny_set_manager.c
 *
 *  \date Apr 11, 2012
 *  \author Alexandre Tuleu
 */

#include <stdlib.h>

#include <sbcpd/error.h>
#include "transfer_manager.h"

//
//inline void increment_tail(sbcpd_tiny_set_manager * tsm){
//	++(tsm->count);
//	++(tsm->tail);
//	if(tsm->tail >= tsm->max_size){
//		tsm->tail = 0;
//	}
//}
//
//inline void increment_head(sbcpd_tiny_set_manager * tsm){
//	++(tsm->head);
//	--(tsm->count);
//	if(tsm->head >= tsm->max_size){
//		tsm->head = 0;
//	}
//}

void sbcpd_init_dl(sbcpd_dl_list * l){
	l->begin = NULL;
	l->end = NULL;
}

void sbcpd_free_dl(sbcpd_dl_list * l){
	while(l->begin){
		sbcpd_dl_item * i = l->begin;
		l->begin = l->begin->next;
		free(i);
	}
}

void sbcpd_dl_push_back(sbcpd_dl_list * l, sbcpd_dl_item * i){
	if(l->end == NULL){
		l->end = i;
		l->begin = i;
		i->prev = NULL;
	} else {
		l->end->next = i;
		i->prev = l->end;
		l->end = i;
	}
	i->next = NULL;
}

sbcpd_dl_item * sbcpd_dl_pop_front(sbcpd_dl_list * l){
	sbcpd_dl_item * i = l->begin;
	if(i == NULL){
		return i;
	}
	l->begin = l->begin->next;
	if(l->begin != NULL){
		l->begin->prev = NULL;
	} else { //there was only one element
		l->end = NULL;
	}
	i->next = NULL;
	//i->prev = NULL;
	return i;
}

int sbcpd_dl_list_empty(sbcpd_dl_list * l) {
	return l->begin == NULL;
}


void sbcpd_dl_remove_item_from_list(sbcpd_dl_list * l , sbcpd_dl_item * i){
	if(i == NULL){
		return;
	}

	if(i->next != NULL){//not at the end of the list
		i->next->prev = i->prev;
	} else {
		if (l->end == i){
			l->end = i->prev;
		}
	}

	if(i->prev != NULL){ //not at the begining of the list
		i->prev->next = i->next;
	} else {
		if( l->begin == i) {
			l->begin = i->next;
		}
	}

}

void sbcpd_tm_init(sbcpd_transfer_manager * tm , uint8_t size){
	if(size == 0){
		return;
	}
	tm->size = size;

	tm->transfers = malloc(sizeof(sbcpd_transfer) * size);
	tm->available_transfers = malloc(sizeof(sbcpd_dl_list));
	tm->allocated_transfers = malloc(sizeof(sbcpd_dl_list));
	sbcpd_init_dl(tm->available_transfers);
	sbcpd_init_dl(tm->allocated_transfers);
	tm->allocated_by_id = malloc(sizeof(sbcpd_dl_item*) * size);

	memset(tm->allocated_by_id,0x00,sizeof(sbcpd_dl_item*) * size);
	unsigned int i;
	for(i = 0 ; i < size ; ++i){
		tm->transfers[i].id = i;
		sbcpd_dl_item * it = malloc(sizeof(sbcpd_dl_item));
		it->t = &(tm->transfers[i]);
		sbcpd_dl_push_back(tm->available_transfers,
		                   it);
	}

	tm->allocated_size = 0;

}

void sbcpd_tm_free(sbcpd_transfer_manager * tm){
	if(tm->transfers){
		free(tm->transfers);
		tm->transfers = NULL;
	}

	if(tm->available_transfers){
		sbcpd_free_dl(tm->available_transfers);
		free(tm->available_transfers);
		tm->available_transfers = NULL;
	}

	if(tm->allocated_transfers){
		sbcpd_free_dl(tm->allocated_transfers);
		free(tm->allocated_transfers);
		tm->allocated_transfers = NULL;
	}


	if(tm->allocated_by_id){
		free(tm->allocated_by_id);
		tm->allocated_by_id = NULL;
	}
}

sbcpd_transfer * sbcpd_tm_allocate_transfer(sbcpd_transfer_manager * tm){
	if(sbcpd_dl_list_empty(tm->available_transfers)){
		return NULL;
	}

	sbcpd_dl_push_back(tm->allocated_transfers,
	                   sbcpd_dl_pop_front(tm->available_transfers));
	sbcpd_transfer * t = tm->allocated_transfers->end->t;
	tm->allocated_by_id[t->id] = tm->allocated_transfers->end;

	++(tm->allocated_size);
	return t;
}

int sbcpd_tm_is_allocated(sbcpd_transfer_manager * tm, sbcpd_transfer * t){
	if(t == NULL){
		return 0;
	}
	return tm->allocated_by_id[t->id] != NULL;
}

void sbcpd_tm_release_transfer(sbcpd_transfer_manager * tm, sbcpd_transfer * t){
	if(!sbcpd_tm_is_allocated(tm,t)){
		return;
	}


	sbcpd_dl_item * item = tm->allocated_by_id[t->id];
	tm->allocated_by_id[t->id] = NULL;

	//remove from allocated.
	sbcpd_dl_remove_item_from_list(tm->allocated_transfers,item);

	//push back to available
	sbcpd_dl_push_back(tm->available_transfers,item);
	--(tm->allocated_size);
}


sbcpd_transfer * sbcpd_tm_first_allocated_transfer(sbcpd_transfer_manager * tm){
	if(tm->allocated_transfers->begin == NULL){
		return NULL;
	}
	return tm->allocated_transfers->begin->t;
}

sbcpd_transfer * sbcpd_tm_next_allocated_transfer(sbcpd_transfer_manager * tm,
                                                  sbcpd_transfer * t){
	if(sbcpd_tm_is_allocated(tm,t) == 0){
		return NULL;
	}

	sbcpd_dl_item * it = tm->allocated_by_id[t->id]->next;
	if(it == NULL){
		return NULL;
	}
	return it->t;
}


uint8_t sbcpd_tm_transfer_count(sbcpd_transfer_manager * tm){
	return tm->allocated_size;
}
