/*
 * Copyright (c) 2011-2013 Alexandre Tuleu <alexandre.tuleu.2005@polytechnique.org>
 * Ecole Polytechnique Fédérale de Lausanne.
 *
 * this file is part of sbcpd.
 *
 * sbcpd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * sbcpd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the Lesser GNU General Public License
 * along with sbcpd.  If not, see <http://www.gnu.org/licenses/>.
 */
/**
 * \file linux_serial.c
 *
 *  \date Apr 17, 2012
 *  \author Alexandre Tuleu
 */

#include <stdlib.h>


#include "linux_serial.h"
#include "linux_serial.ih"

#include <fcntl.h>

#include <math.h>

sbcpd_serial_interface * sbcpd_linux_serial_create_interface(){
	sbcpd_serial_interface * si = malloc(sizeof(sbcpd_serial_interface));

	sbcpd_linux_serial * ldev = malloc(sizeof(sbcpd_linux_serial));
	ldev->fd = -1;
	si->dev = ldev;
	si->free = &sbcpd_linux_serial_free;
	si->close = &sbcpd_linux_serial_close;
	si->write_data = &sbcpd_linux_serial_write;
	si->read_data = &sbcpd_linux_serial_read;

	return si;
}

int sbcpd_linux_serial_open(sbcpd_serial_interface * si,
                            const char * devname,
                            unsigned int baudrate){
	sbcpd_linux_serial * ldev = si->dev;
	ldev->fd = open(devname, O_RDWR | O_NOCTTY | O_NONBLOCK);
	if(ldev->fd < 0 ){
		SBCPD_RETURN_WITH_ERROR(si,
		                        SBCPD_ESI_CANNOT_ACCESS_DEVICE,
		                        "Could not open device.");
	}

	if(tcgetattr(ldev->fd,&(ldev->state)) < 0){
		sbcpd_linux_serial_close(si);
		SBCPD_RETURN_WITH_ERROR(si,
		                        SBCPD_ESI_CONFIGURE_ERROR,
		                        "Could not open device state.");
	}

	//copied config from DMP's RoBoardIO library v1.8.1
	ldev->state.c_cflag |= CLOCAL | CREAD;
	ldev->state.c_cflag &= ~CRTSCTS;
	ldev->state.c_lflag &= (unsigned int)~(ICANON | ISIG | IEXTEN | ECHO | ECHOE);
	ldev->state.c_iflag &= (unsigned int)~OPOST;
	ldev->state.c_cc[VTIME] = 0;
	ldev->state.c_cc[VMIN] = 0;

	//set 8 Bit per Byte 1 stop byte no parity
	ldev->state.c_cflag = (ldev->state.c_cflag & (unsigned int)~CSIZE) | CS8;
	ldev->state.c_cflag &= (unsigned int)~(CSTOPB); // 1 stop bit
	//no parity
	ldev->state.c_cflag &= (unsigned int)~PARENB;
	ldev->state.c_iflag &= (unsigned int)~(INPCK | ISTRIP | PARMRK);


	const static unsigned int possible_baudrate[10] = {50,300,1200,2400,4800,9600,19200,
	                                          38400,57600,115200};
	const static speed_t possible_baudrate_[10] = {B50,B300,B1200,B2400,B4800,
	                                               B9600,B19200,B38400,B57600,
	                                               B115200};

	unsigned int i;
	int chosen = -1;
	for( i = 0; i < 10; ++i){
		float error = fabs(possible_baudrate[i] - baudrate) / possible_baudrate[i];
		if(error < 0.03){
			chosen = (int)i;
			break;
		}
	}

	if(chosen == -1){
		sbcpd_linux_serial_close(si);
		SBCPD_RETURN_WITH_ERROR(si,SBCPD_ESI_CONFIGURE_ERROR,
		                        "Unsupported baudrate.");
	}

	cfsetispeed(&(ldev->state),possible_baudrate_[chosen]);
	cfsetospeed(&(ldev->state),possible_baudrate_[chosen]);


	if(tcsetattr(ldev->fd,(size_t)TCSANOW,&(ldev->state)) < 0){
		sbcpd_linux_serial_close(si);
		SBCPD_RETURN_WITH_ERROR(si,SBCPD_ESI_CONFIGURE_ERROR,
		                        "Could not configure device mode.");
	}

	tcflush(ldev->fd,(size_t)TCIOFLUSH);

	return SBCPD_NO_ERROR;
}


void sbcpd_linux_serial_close(sbcpd_serial_interface * si){
	sbcpd_linux_serial * ldev = si->dev;
	if(ldev->fd != -1){
		close(ldev->fd);
	}
	ldev->fd = -1;
}

void sbcpd_linux_serial_free(sbcpd_serial_interface * si){
	sbcpd_linux_serial_close(si->dev);
	free(si->dev);
}

//it seems that we should write in chunk of 256 max

int sbcpd_linux_serial_write(sbcpd_serial_interface * si,
                             const sbcpd_byte_t * data,
                             int size){
	sbcpd_linux_serial * ldev = si->dev;
	tcflush(ldev->fd,TCIFLUSH);
	return write(ldev->fd,data,(size_t)size);
}

int sbcpd_linux_serial_read(sbcpd_serial_interface * si,
                            sbcpd_byte_t * data,
                            int size){
	sbcpd_linux_serial * ldev = si->dev;
	return read(ldev->fd,data,(size_t)size);
}
