/*
 * Copyright (c) 2013 Alexandre Tuleu <alexandre.tuleu.2005@polytechnique.org>
 * Ecole Polytechnique Fédérale de Lausanne.
 *
 * this file is part of sbcpd.
 *
 * sbcpd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * sbcpd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the Lesser GNU General Public License
 * along with sbcpd.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * \file stats.c
 *
 *  \date May 20, 2013
 *  \author Alexandre Tuleu
 */

#include "stats.h"
#include <sbcpd/private/stats.h>

#include <stdlib.h>
#include <string.h>

sbcpd_stats * sbcpd_create_stats(){
	sbcpd_stats * stats = malloc(sizeof(sbcpd_stats));

	memset(stats,0,sizeof(sbcpd_stats));

	return stats;
}

void sbcpd_free_stats(sbcpd_stats * s){
	free(s);
}

void sbcpd_stats_log_transfer(sbcpd_stats * s, sbcpd_transfer * t){
	s->total_packet += 1;
	if ( t->status == SBCPD_TRANSFER_OK) {
		return;
	}

	if (t->status == SBCPD_TRANSFER_HAS_APPLICATION_ERROR){
		s->app_errors += 1;
		return;
	}

	if(t->status == SBCPD_TRANSFER_HAS_COMMUNICATION_ERROR){
		s->com_errors[SBCPD_CERR_NO_ERROR] += 1;
		s->com_errors[t->com_error] += 1;
		return;
	}

}


sbcpd_nb_packet_t sbcpd_stats_get_nb_packet(sbcpd_stats * s){
	return s->total_packet;
}

sbcpd_nb_packet_t sbcpd_stats_get_slave_application_errors(sbcpd_stats * s){
	return s->app_errors;
}

sbcpd_nb_packet_t sbcpd_stats_get_nb_com_errors(sbcpd_stats * s,
                                                sbcpd_transfer_communication_error type){
	if (type < SBCPD_CERR_NO_ERROR || type > SBCPD_CERR_NUMBER_OF_ERRORS){
		return 0;
	}
	return s->com_errors[type];
}

void sbcpd_stats_fprintf(FILE * f, sbcpd_stats * s){
	if(s->total_packet == 0) {
		fprintf(f,"No packet have transited over the bus yet.");
	}
	double success = s->total_packet - s->com_errors[SBCPD_CERR_NO_ERROR] - s->app_errors;
	success /= s->total_packet;
	success *= 100.0;
	double app_error = 100.0 * ((double) s->app_errors) / ( (double) s->total_packet);
	double com_error = 100.0 - success - app_error;
	fprintf(f,
			"%llu packet transmitted (Success: %f %% Communication Error: %f %% Application Error. %f %%",
			(long long unsigned)(s->total_packet),
			success,
			com_error,
			app_error);
}

#define fprintf_stats_detail(f,s,cerr) do{ \
	fprintf(f,"\n * %s : %f %%", #cerr , 100.0 * ((double)s->com_errors[cerr])/((double)s->com_errors[SBCPD_CERR_NO_ERROR])); \
}while(0)

void sbcpd_stats_detailed_fprintf(FILE * f, sbcpd_stats * s){
	sbcpd_stats_fprintf(f,s);

	if(s->total_packet == 0){
		return;
	}
	if(s->com_errors[SBCPD_CERR_NO_ERROR] == 0){
		return;
	}
	fprintf(f,"\ndetails of Communication errors :");
	fprintf_stats_detail(f,s,SBCPD_CERR_BAD_CERR_PACKET);
	fprintf_stats_detail(f,s,SBCPD_CERR_IMCOMPLETE_READOUT);
	fprintf_stats_detail(f,s,SBCPD_CERR_INCONSISTENT_RESPONSE_SIZE);
	fprintf_stats_detail(f,s,SBCPD_CERR_HOST_FRAME_ERROR);
	fprintf_stats_detail(f,s,SBCPD_CERR_MASTER_FRAME_ERROR);
	fprintf_stats_detail(f,s,SBCPD_CERR_SLAVE_FRAME_ERROR);
	fprintf_stats_detail(f,s,SBCPD_CERR_HOST_TIMEDOUT);
	fprintf_stats_detail(f,s,SBCPD_CERR_MASTER_TIMEDOUT);
	fprintf_stats_detail(f,s,SBCPD_CERR_SLAVE_TIMEOUT);
	fprintf_stats_detail(f,s,SBCPD_CERR_HOST_BAD_CHECKSUM);
	fprintf_stats_detail(f,s,SBCPD_CERR_MASTER_BAD_CHECKSUM);
	fprintf_stats_detail(f,s,SBCPD_CERR_SLAVE_BAD_CHECKSUM);
	fprintf_stats_detail(f,s,SBCPD_CERR_UNKNOWN_INSTRUCTION_ERROR);

}

