/*
 * Copyright (c) 2011-2013 Alexandre Tuleu <alexandre.tuleu.2005@polytechnique.org>
 * Ecole Polytechnique Fédérale de Lausanne.
 *
 * this file is part of sbcpd.
 *
 * sbcpd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * sbcpd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the Lesser GNU General Public License
 * along with sbcpd.  If not, see <http://www.gnu.org/licenses/>.
 */
/**
 * \file BusUTest.cpp
 *
 *  \date Apr 16, 2012
 *  \author Alexandre Tuleu
 */

#include "BusUTest.h"

#include <sbcpd/settings.h>
#include <sbcpd/sbcp.h>

#include <sbcpd/bus.ih>

using namespace ::testing;

sbcpd_byte_t * BusUTest::ExpectMessageExchange(sbcpd_packet_size_t out_size,
                                               sbcpd_packet_size_t in_size,
                                               sbcpd_byte_t error){
	InSequence messageStack;
	sbcpd_byte_t * data  =mock_si->AppendResponse(in_size);
	data[SBCP_HEADER_POS] = SBCP_HEADER_VALUE;
	data[SBCP_CLASS_POS]  = 0x00;
	data[SBCP_ID_POS]     = 0x00;
	data[SBCP_INSTRUCTION_POS]  = error;
	data[SBCP_PAYLOAD_SIZE_POS] = in_size - SBCP_MINIMAL_SIZE;

	EXPECT_CALL(*mock_si,Write(::testing::_,out_size)).WillOnce(Return(out_size)).RetiresOnSaturation();
	EXPECT_CALL(*mock_si,Read(::testing::_,::testing::_)).WillOnce(Return(in_size)).RetiresOnSaturation();

	return data;
}



void BusUTest::SetUp(){
	mock_si = new MockSerialInterface();
	sbcpd_bus_settings * s = sbcpd_bus_settings_create();

	bus = sbcpd_create_bus(s);

	sbcpd_byte_t * d = ExpectMessageExchange(6,8,SBCP_NO_ERROR);
	d[SBCP_PAYLOAD_START_POS]     = 0x00;
	d[SBCP_PAYLOAD_START_POS + 1] = 0x01;
	d[SBCP_PAYLOAD_START_POS + 2] = sbcpd_compute_checksum_of_packet(d);

	d = ExpectMessageExchange(8,10,SBCP_NO_ERROR);
	d[SBCP_PAYLOAD_START_POS]     = 0x00;
	d[SBCP_PAYLOAD_START_POS + 1] = 8;
	d[SBCP_PAYLOAD_START_POS + 2] = 0;
	d[SBCP_PAYLOAD_START_POS + 3] = 64;
	d[SBCP_PAYLOAD_START_POS + 4] = sbcpd_compute_checksum_of_packet(d);

	ASSERT_EQ(SBCPD_NO_ERROR,sbcpd_bus_open(bus,mock_si->SI()))
	 << "Got error on opening bus : " << sbcpd_bus_get_last_error_message(bus);
	t1=NULL;
	t2=NULL;
	t3=NULL;
}

void BusUTest::TearDown(){
	if(bus->flags & SBCPD_BUS_OPEN){
		EXPECT_CALL(*mock_si,Close());
		EXPECT_CALL(*mock_si,Free());
		sbcpd_bus_close(bus);
	}
	//only the sbcpd_serial_interface * is managed. we have to delete our mock
	//manually
	delete mock_si;
}


void UnMasterisedBusUTest::SetUp(){
	mock_si = new MockSerialInterface();
	sbcpd_bus_settings * s = sbcpd_bus_settings_create();
	sbcpd_bus_settings_set_required_frame_definition(s,1,10);
	bus = sbcpd_create_bus(s);

	mock_si->AppendResponse(0);
	EXPECT_CALL(*mock_si,Write(_,6)).WillOnce(Return(6));
	EXPECT_CALL(*mock_si,Read(_,_)).WillRepeatedly(Return(0));

	EXPECT_EQ(SBCPD_NO_ERROR,sbcpd_bus_open(bus,mock_si->SI()))
		<< "Got error on opening bus : " << sbcpd_bus_get_last_error_message(bus);
	t = NULL;
}

void UnMasterisedBusUTest::TearDown(){
	EXPECT_CALL(*mock_si,Close());
	EXPECT_CALL(*mock_si,Free());
	sbcpd_bus_close(bus);
	delete mock_si;
}

TEST_F(BusUTest,NewlyOpenedBusHaveNoPendingTransfer){
	EXPECT_TRUE(sbcpd_bus_frame_is_empty(bus));
}

TEST_F(BusUTest,TransferCouldBeReleased){
	sbcpd_transfer * t = sbcpd_bus_append_transfer(bus,0);

	EXPECT_EQ(1,sbcpd_bus_current_frame_size(bus));

	sbcpd_bus_release_transfer(bus,t);

	EXPECT_TRUE(sbcpd_bus_frame_is_empty(bus));
}


TEST_F(BusUTest,FrameCouldBeEmptied){
	PrepareFrame();
	EXPECT_EQ(3,sbcpd_bus_current_frame_size(bus));

	EXPECT_EQ(SBCPD_NO_ERROR, sbcpd_bus_release_all_transfers(bus)) << "Got Deallocation error : " << sbcpd_bus_get_last_error_message(bus);

	EXPECT_TRUE(sbcpd_bus_frame_is_empty(bus));
}

void BusUTest::PrepareFrame(){
	t1 = sbcpd_bus_append_transfer(bus,0);
	ASSERT_NE((sbcpd_transfer*)NULL,t1) << "Got allocation error : "
			<< sbcpd_bus_get_last_error_message(bus);

	t2 = sbcpd_bus_append_transfer(bus,0);
	ASSERT_NE((sbcpd_transfer*)NULL,t2) << "Got allocation error : "
			<< sbcpd_bus_get_last_error_message(bus);

	t3 = sbcpd_bus_append_transfer(bus,0);
	ASSERT_NE((sbcpd_transfer*)NULL,t3) << "Got allocation error : "
			<< sbcpd_bus_get_last_error_message(bus);
}

TEST_F(BusUTest,FrameCouldBeCompromised){
	PrepareAndCompromiseFrame();
	EXPECT_FALSE(sbcpd_bus_good(bus));

	EXPECT_EQ(NULL,sbcpd_bus_append_transfer(bus,0));

	EXPECT_EQ(SBCPD_EBUS_FRAME_COMPROMISED,sbcpd_bus_get_last_error_code(bus));
}


void BusUTest::CompromiseFrame(){
	ASSERT_EQ(SBCPD_NO_ERROR,sbcpd_bus_release_transfer(bus,t2)) <<
	"Got Deallocation error : " << sbcpd_bus_get_last_error_message(bus);

}

void BusUTest::PrepareAndCompromiseFrame(){
	PrepareFrame();
	CompromiseFrame();
}

void BusUTest::PrepareTimeoutByMaster(){
	EXPECT_CALL(*mock_si,Write(_,18));
	sbcpd_byte_t * data = mock_si->AppendResponse(18);
	for(unsigned int i = 0; i < 3; ++i){
		data[6 * i + SBCP_HEADER_POS] = SBCP_HEADER_VALUE;
		data[6 * i + SBCP_CLASS_POS] = SBCP_MASTER_CLASS;
		data[6 * i + SBCP_ID_POS] = 0x42;
		data[6 * i + SBCP_INSTRUCTION_POS] = SBCP_CERR_SLAVE_TIMEOUT;
		data[6 * i + SBCP_PAYLOAD_SIZE_POS] = 0;
		data[6 * i + SBCP_PAYLOAD_START_POS] = sbcpd_compute_checksum_of_packet(&(data[6 * i]));
	}

	EXPECT_CALL(*mock_si,Read(_,_)).WillOnce(Return(21));
}

TEST_F(BusUTest,FrameCouldAlwaysBeResetted){
	PrepareAndCompromiseFrame();
	//now frame is compromised

	//case 1 : bus not transmitting.
	EXPECT_EQ(SBCPD_NO_ERROR,sbcpd_bus_release_all_transfers(bus));

	EXPECT_TRUE(sbcpd_bus_good(bus));

	PrepareAndCompromiseFrame();
	//case 2 : we start the transfer after compromisation
	EXPECT_EQ(SBCPD_EBUS_FRAME_COMPROMISED,sbcpd_bus_submit_all_transfer(bus));

	EXPECT_EQ(SBCPD_NO_ERROR,sbcpd_bus_release_all_transfers(bus));

	EXPECT_TRUE(sbcpd_bus_good(bus));


	//case 3 : we release after waiting for IO.
	PrepareAndCompromiseFrame();

	EXPECT_EQ(SBCPD_EBUS_FRAME_COMPROMISED,sbcpd_bus_submit_all_transfer(bus));

	EXPECT_EQ(SBCPD_EBUS_FRAME_COMPROMISED,sbcpd_bus_wait_for_completion(bus,-1));

	EXPECT_EQ(SBCPD_NO_ERROR,sbcpd_bus_release_all_transfers(bus));

	EXPECT_TRUE(sbcpd_bus_good(bus));



	//case 4 : we release after submission and before termination.

	PrepareFrame();

	PrepareTimeoutByMaster();

	EXPECT_EQ(SBCPD_NO_ERROR,sbcpd_bus_submit_all_transfer(bus)) << "Got submission error : " << sbcpd_bus_get_last_error_code(bus);

	//here the test is compromised since completion does happen inside submission.

	CompromiseFrame();

	EXPECT_EQ((sbcpd_transfer*)NULL,sbcpd_bus_append_transfer(bus,0));
	EXPECT_EQ(SBCPD_EBUS_FRAME_COMPROMISED,sbcpd_bus_get_last_error_code(bus));

	EXPECT_EQ(SBCPD_EBUS_FRAME_COMPROMISED,sbcpd_bus_wait_for_completion(bus,-1));

	EXPECT_EQ(SBCPD_NO_ERROR,sbcpd_bus_release_all_transfers(bus));

	EXPECT_TRUE(sbcpd_bus_good(bus));

}

TEST_F(BusUTest,CouldRecoverFromAFrameReadingCrash){

	unsigned int frame_size = sbcpd_bus_max_packet_per_frame(bus);
	sbcpd_transfer ** t = new sbcpd_transfer*[frame_size];
	for(unsigned int i = 0; i < frame_size; ++i){
		t[i] = sbcpd_bus_append_transfer(bus,0);
		ASSERT_NE((sbcpd_transfer*)NULL,t[i]);
	}

	mock_si->AppendResponse(0);//nobody will respond at all ... crash will be read

	EXPECT_CALL(*mock_si,Write(_,frame_size*6)).WillOnce(Return(8*6));

	//now we sets that the driver will always return no byte read, without problem
	EXPECT_CALL(*mock_si,Read(_,_)).WillRepeatedly(Return(0));

	EXPECT_EQ(SBCPD_NO_ERROR,sbcpd_bus_submit_all_transfer(bus));

	EXPECT_EQ(SBCPD_NO_ERROR,sbcpd_bus_wait_for_completion(bus,-1));

	for(unsigned int i = 0; i < frame_size; ++i){
		EXPECT_EQ(SBCPD_TRANSFER_HAS_COMMUNICATION_ERROR,t[i]->status);
		EXPECT_EQ(SBCPD_CERR_IMCOMPLETE_READOUT,t[i]->com_error);
	}

	EXPECT_FALSE(sbcpd_bus_is_busy(bus));

	EXPECT_EQ(SBCPD_NO_ERROR,sbcpd_bus_release_all_transfers(bus));

	EXPECT_EQ(0,sbcpd_bus_current_frame_size(bus));

	EXPECT_NE((sbcpd_transfer*)NULL,sbcpd_bus_append_transfer(bus,0))
		<< "Allocation problem : " << sbcpd_bus_get_last_error_message(bus);
	delete [] t;
}


TEST_F(UnMasterisedBusUTest,CouldRecoverFromReadCrash){
	t = sbcpd_bus_append_transfer(bus,0);

	EXPECT_CALL(*mock_si,Write(_,6)).WillOnce(Return(6));
	EXPECT_CALL(*mock_si,Read(_,_)).WillRepeatedly(Return(0));

	EXPECT_EQ(SBCPD_NO_ERROR,sbcpd_bus_submit_all_transfer(bus));

	EXPECT_EQ(SBCPD_NO_ERROR,sbcpd_bus_wait_for_completion(bus,-1));

	EXPECT_EQ(SBCPD_TRANSFER_HAS_COMMUNICATION_ERROR,t->status);
	EXPECT_EQ(SBCPD_CERR_IMCOMPLETE_READOUT,t->com_error);

	EXPECT_FALSE(sbcpd_bus_is_busy(bus));

	EXPECT_EQ(SBCPD_NO_ERROR,sbcpd_bus_release_all_transfers(bus));

	EXPECT_EQ(0,sbcpd_bus_current_frame_size(bus));

	EXPECT_TRUE(sbcpd_bus_good(bus));

	EXPECT_NE((sbcpd_transfer*)NULL, sbcpd_bus_append_transfer(bus,0))
		<< "Allocation problem : " << sbcpd_bus_get_last_error_message(bus);


}
