/*
 * Copyright (c) 2011-2013 Alexandre Tuleu <alexandre.tuleu.2005@polytechnique.org>
 * Ecole Polytechnique Fédérale de Lausanne.
 *
 * this file is part of sbcpd.
 *
 * sbcpd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * sbcpd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the Lesser GNU General Public License
 * along with sbcpd.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * \file transfer.h
 *
 * \date Aug 15, 2011
 * \author Alexandre Tuleu
 */

#ifndef SBCPD_TRANSFER_H_
#define SBCPD_TRANSFER_H_

#include <stdio.h>

#include "common.h"

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Transfer status.
 * \ingroup sbcpd_transfer_m
 */
enum sbcpd_transfer_status{
	SBCPD_TRANSFER_OK = 0x00,
	//!< SBCPD_TRANSFER_OK the transfer is finished and good.
	SBCPD_TRANSFER_UNSTARTED,
	//!< SBCPD_TRANSFER_UNSTARTED the transfer is not started yet, if you get
	//!< this state in your callback function, this is a bug the high level
	//!< driver.
	SBCPD_TRANSFER_IN_PROGRESS,
	//!< SBCPD_TRANSFER_IN_PROGRESS the transfer is started and the driver is
	//!< waiting its completion. If you get it in your callback function, then
	//!< it is a bug of the high level driver.
	SBCPD_TRANSFER_ABORTED,
	//!< SBCPD_TRANSFER_ABORTED the current transfer is aborted.
	SBCPD_TRANSFER_HAS_COMMUNICATION_ERROR,
	//!< SBCPD_TRANSFER_HAS_COMMUNICATION_ERROR the received packet has some
	//!< communication errors.
	SBCPD_TRANSFER_HAS_APPLICATION_ERROR
	//!< SBCPD_TRANSFER_HAS_APPLICATION_ERROR the received packet indicates an
	//!< application-specific error
};
typedef enum sbcpd_transfer_status sbcpd_transfer_status;

enum sbcpd_transfer_communication_error{
	SBCPD_CERR_NO_ERROR = 0x00,
	//!< SBCPD_CERR_NO_ERROR No communication error
	SBCPD_CERR_BAD_CERR_PACKET,
	//!< SBCPD_CERR_BAD_CERR_PACKET The packet issuing a communication error
	//!< is badly formatted.
	SBCPD_CERR_IMCOMPLETE_READOUT,
	//!< SBCPD_CERR_IMCOMPLETE_READOUT the driver did not get enough data
	//!< from the bus to read any response.
	SBCPD_CERR_INCONSISTENT_RESPONSE_SIZE,
	//!< SBCPD_CERR_INCONSISTENT_RESPONSE_SIZE the size of the response is
	//!< is bad
	SBCPD_CERR_HOST_FRAME_ERROR,
	//!< SBCPD_CERR_HOST_FRAME_ERROR the master got a frame error from the
	//!< host
	SBCPD_CERR_MASTER_FRAME_ERROR,
	//!< SBCPD_CERR_MASTER_FRAME_ERROR a slave got a frame error from the
	//!< master
	SBCPD_CERR_SLAVE_FRAME_ERROR,
	//!< SBCPD_CERR_SLAVE_FRAME_ERROR the master of the bus got an UART frame
	//!< error from a slave.
	SBCPD_CERR_HOST_TIMEDOUT,
	//!< SBCPD_CERR_HOST_TIMEDOUT the master of the bus issued a timeout
	//!< while receiving an host packet
	SBCPD_CERR_MASTER_TIMEDOUT,
	//!< SBCPD_CERR_MASTER_TIMEDOUT a slave of the bus issued a timeout
	//!< while reciving a master packet.
	SBCPD_CERR_SLAVE_TIMEOUT,
	//!< SBCPD_CERR_SLAVE_TIMEOUT the master of the bus issued a timeout
	//!< while waiting for a slave response.
	SBCPD_CERR_HOST_BAD_CHECKSUM,
	//!< SBCPD_CERR_HOST_BAD_CHECKSUM the master of the bus issued a bad
	//!< checksum while receiving an host packet.
	SBCPD_CERR_MASTER_BAD_CHECKSUM,
	//!< SBCPD_CERR_MASTER_BAD_CHECKSUM a slave of the bus got issued a bad
	//!< checksum while receiving a packet from master.
	SBCPD_CERR_SLAVE_BAD_CHECKSUM,
	//!< SBCPD_CERR_SLAVE_BAD_CHECKSUM the data recived back from the master
	//!< does not have a correct checksum.
	SBCPD_CERR_UNKNOWN_INSTRUCTION_ERROR,
	//!< SBCPD_CERR_UNKNOWN_INSTRUCTION_ERROR a slave did not recognize
	//!< the asked instruction.
	SBCPD_CERR_UNEXPECTED_SIZE,
	SBCPD_CERR_REG_NOT_READABLE,
	SBCPD_CERR_REG_NOT_WRITABLE,
	SBCPD_CERR_REG_PROTECTED,
	SBCPD_CERR_NUMBER_OF_ERRORS
};

typedef enum sbcpd_transfer_communication_error
	sbcpd_transfer_communication_error;


/**
 * \brief Callback which will be used once a transfer is finished.
 * \param a pointer to the sbcpd_transfer structure that finished.
 * \ingroup sbcpd_transfer_m
 */
typedef void (*sbcpd_transfer_callback)(sbcpd_transfer*);



/**
 * A structure that represent a transfer over a SBCP bus. Please note that this
 * request should not be allocated by yourself using malloc, but only with
 * sbcpd_create_transfer().
 * \ingroup sbcpd_transfer_m
 */
struct sbcpd_transfer_str {
	uint8_t                            id;
	sbcpd_transfer_status              status;
	sbcpd_transfer_communication_error com_error;
	sbcpd_transfer_callback            callback;
	sbcpd_byte_t * const             out_buffer;
	const sbcpd_byte_t * const       in_buffer;
	const unsigned int                 out_size;
	const unsigned int                 in_size;
	void *                             user_data;
};


const char * sbcpd_transfer_status_msg(sbcpd_transfer_status s);

void sbcpd_transfer_fprintf(sbcpd_transfer * t, FILE * f);


uint8_t sbcpd_transfer_get_uuid(sbcpd_transfer * t);

/**
 * \defgroup sbcpd_transfer_m SBCP Transfer
 * This module exposes a set of function and structire that should be used to
 * send and receive packet over an opened sbcpd_bus. SBCP bus have rather
 * limited ressources, and thus you can only have a relatively small number of
 * concurrent SBCP transfer (4 at this time per bus). Therefore transfer request
 * allocated with sbcpd_create_transfer() should be released as soon as possible
 * with sbcpd_free_transfer().
 *
 *
 * Basically, here are the operation needed to start any transfer over an
 * initialized bus :
 *  - Allocate an sbcpd_transfer
 *  - Initialize any data that is wanted to be sent over the bus
 *  - Prepare a callback function that will be used when feedback packet will
 *    be received.
 *  - Submit the transfer to the bus.
 *  - Free the sbcpd_transfer
 *
 * In order to ensure the lowest latency over the SBCP bus, their is one major
 * limitation when manipulating transfers : actual transfer of any data over the
 * bus will not take place until all allocated request are submitted. Therefore
 * their one can see two good practice to manage its transfers :
 *
 * - One shot transfer : These tansfers are only used once. They should free the
 *   sbcpd_transfer in their request
 * - Continuous transfer : These transfer are polled periodically. They should
 *   re submit the sbcpd_transfer in their callback function.
 *
 * Please notice that all execution during a callback function should be as
 * minimal as possible. During the process of a callback function, all sending
 * and receiving of the bus is suspended.
 *
 *
 *
 * \ingroup sbcpd_bus_m
 */

#ifdef __cplusplus
}
#endif

#endif //SBCPD_TRANSFER_H_
